import numpy as np
from scipy.special import lambertw

def COP(x,a1,a2,tol=1e-5):
    """
    Dimensionless Chu-Owens-Peters model for burst nucleation, y = a1*(sqrt(1 + 4*a2/a1*x) - 1)
    
        Parameters:
            x (array_like): The independent variable (solute generation rate).
            a1 (float): Critical nucleus size parameter.
            a2 (float): Growth parameter.
            
        Returns:
            y (array_like): The dependent variable of the data (final number of nuclei).
    """
    alph = np.atleast_1d(a2*x/a1)
    flag = alph > tol
    if flag.all():
        y = a1*(np.sqrt(1+4*alph)-1)
    elif np.invert(flag).all():
        y = 2*a1*alph*(1-alph)
    else:
        s = alph.shape
        y = np.atleast_1d(np.zeros(s))
        if np.shape(a1) != s:
            a1 = a1*np.ones(s)
        if np.shape(a2) != s:
            a2 = a2*np.ones(s)
        if np.shape(x) != s:
            x = x*np.ones(s)
        y[flag] = a1[flag]*(np.sqrt(1+4*alph[flag])-1)
        flag = np.invert(flag)
        y[flag]=2*a1[flag]*alph[flag]*(1-alph[flag])
    return y


def log_COP(x,b1,b2):
    alph = np.atleast_1d(b2 - b1 + x)
    flag = alph > -15
    if flag.all():
        y = b1 + np.log(np.sqrt(1+4*np.exp(alph))-1)
    else:
        s = alph.shape
        y = np.atleast_1d(np.zeros(s))
        if np.shape(b1) != s:
            b1 = b1*np.ones(s)
        if np.shape(b2) != s:
            b2 = b2*np.ones(s)
        if np.shape(x) != s:
            x = x*np.ones(s)
        y[flag] = np.atleast_1d(b1)[flag] + np.log(np.sqrt(1+4*np.exp(alph[flag]))-1)
        flag = np.invert(flag)
        y[flag]=np.log(2) + np.atleast_1d(b2)[flag] + np.atleast_1d(x)[flag] - np.exp(alph[flag])
    return y


def S8_abs(t,k,kp):
    """
    Compute the S8 concentration/absorbance early in the reaction.
    
        Parameters:
            t (array_like): The independent variable (time).
            k (float): Final rate constant.
            k1 (float): Multiplicative rate factor.
        
        Returns:
            y (array_like): The concentration/absorbance.
    """
    y = kp*(t+(np.exp(-k*t)-1)/k)
    return y


def rf_q0(x,K2):
    """
    Compute the rate factor q0 =HT^2*(K2+T)/(K2+H+T)^2, such that S(t) = k'q0(t + (exp(-kt)-1)/k).
    
        Parameters:
            x (array_like): Array where x[0] = H, x[1] = T
            K2 (float): First rate constant.
        
        Returns:
            y (array_like): The concentration/absorbance.
    """
    H,T = x
    q0 = H*(T**2)*(K2+T)/(K2+H+T)**2
    return q0   


def ktm(x,k,kp,K2):
    q0 = rf_q0(x,K2)
    a = 1 + k/(kp*q0)
    y = a + np.real(lambertw(-np.exp(-a)))
    return y

