import numpy as np
import os
import pandas as pd
import re

## mathematical tools
def resid(p,f,x,y,weights=None):
    """
    Returns the residuals of xy data and a function with optional weights.
    
        Parameters:
            p (array_like or float): Fit parameters to be passed to the function.
            f (callable): Function f(x,*p) which predicts the dependent variable.
            x (array_like): The independent variable of the data.
            y (array_like): The dependent variable of the data.
            weights (array_like): Variance weights of the data points.
        
        Returns:
            resid (array_like): Residual between prediction and true values of the data.
    """
    yhat = f(x,*p)
    res = yhat - y
    if weights:
        res /= np.sqrt(weights)
    return res
    
    
def jac(f,x,p,s=1e-6):
    """
    Computes the Jacobian of the function(s) f(x,p) w.r.t. p using finite differences
    
        Parameters:
            f (callable): Function(s) f(x,*p) which predicts the dependent variable.
            x (array_like): The independent variable.
            p (array_like): Parameters to be passed to the function.
            s (float): scale factor for finite difference step size
        
        Returns:
            J (array_like): Jacobian/gradient
    """
    yhat = np.array(f(x,*p),dtype=float)
    P = len(p)
    J = np.zeros(yhat.shape+(P,))  # gradient extends dimension
    z = np.zeros(P)
    for i in range(P):
        h = (abs(p[i]) + s**2)*s
        dp = z.copy()
        dp[i] = h
        df = np.subtract(f(x,*(p + dp)),f(x,*(p - dp)))
        J[...,i] = 0.5*df/h
    return J
    

## file system tools
def ensure_dir(dirpath):
    """ Ensure that a directory exists. """
    if not os.path.isdir(dirpath):
        os.makedirs(dirpath)
    return
    

## data tools
def extract_nums(s,idxs='all'):
    """
    Returns the numerical entries in a string.
    
        Parameters:
            s (string): String to search for numerical entries.
            idxs (int): Index (or list of indices) of match(es) to return.
        
        Returns:
            matches
    """
    matches = re.findall('\d*\.?\d+',s)
    if len(matches) == 0:
        return np.nan
    elif idxs == 'all':
        return matches
    else:
        return matches[idxs]

def dfappend(df,data):
    datdf = pd.DataFrame(data,columns=df.columns)
    df = df.append(datdf)
    return df
