import numpy as np
import matplotlib.pyplot as plt
from src.tools import jac
# from matplotlib.ticker import *

def datplot(x,y,ax=None,log=[0,0],**kwargs):
    """
    Creates a plot with error bars for discrete data.
    
        Parameters:
            x (array_like): Values of the data for the abscissa.
            y (array_like): Values of the data for the oordinate.
            ax (Axes): Optional existing axes on which to plot.
            log (bool): Flags whether or not to plot log scale (currently not supported for error bars).
        
        Returns:
            ax (Axes): Final axes object containing the plot.
            handles (list): Handles of the plotted elements.
            
        Other Parameters:
            **kwargs (dict): All other keyword arguments passed on to the pyplot call.
    """
    ## do log plotting if applicable
    dat = [x,y]
    dat = [np.log(dat[i]) if log[i] else dat[i] for i in range(2)]
    ## make axes if none given
    errbarargs = {'fmt':'o','elinewidth':1,'capsize':2}
    errbarargs.update(kwargs)
    if ax == None:
        fig, ax = plt.subplots()
    handles = ax.errorbar(*dat,**errbarargs)
    
    plt.close()
    
    return ax, handles

def funplot(f,p,p_err=None,n=2,ax=None,xs=None,alpha=0.1):
    """
    Creates a plot with error bars for a function.
    
        Parameters:
            f (callable): Function f(xs,*p) which predicts the dependent variable.
            p (array_like or float): Fit parameters to be passed to the function.
            p_err (array_like or float): Error on the parameters to be passed to the function.
            n (int or list): Number of standard error regions to draw or values of standard errors at which to draw lines.
            ax (Axes): Optional existing axes on which to plot.
            xs (array_like): The limits or values of the independent variable at which to plot.
        
        Returns:
            ax (Axes): Final axes object containing the plot.
            handles (list): Handles of the plotted elements.
    """
    ## make axes if none given
    if ax == None:
        fig, ax = plt.subplots()
    ## determine plotting points
    if np.array(xs == None).any():
        xs = np.linspace(*ax.get_xlim())
        ax.set_autoscalex_on(False)
    elif len(xs) == 2:
        ax.set_xlim(xs)
        xs = np.linspace(*xs)
        
    ## make sure p is an array        
    p = np.atleast_1d(p)
    
    handles = ()
    ## plot line of function
    ys = f(xs,*p)
    h = ax.plot(xs,ys)
    handles += (h[0],)
    if np.array(n).size == 1:
        sig_ns = range(1,n+1)
    else:
        sig_ns = n
    ## make sure p_err is an array
    p_err = np.atleast_1d(p_err)
    ## if errors are present
    if (np.atleast_1d(p_err)!=None).all():
        ## get color of line
        c = ax.lines[-1].get_color()
        ## ensure matrix
        if len(p_err.shape) == 1:
            p_err = np.diag(p_err)
        J = jac(f,xs,p)
        dy = np.sqrt(np.diag(J @ p_err @ J.T))
        for i in sig_ns:
            ## make plots
            h = ax.fill_between(xs,ys + i*dy,ys - i*dy,alpha=alpha,color = c)
            handles += (h,)
    
    plt.close()
    return ax, handles


def contour2d(f,x,y,p=[],ax=None,log=[0,0,0],filled=True,new_cbar=True,**kwargs):
    """
    Creates a filled contour plot of a function with a color bar.
    
        Parameters:
            f (callable): Function f(x,y,*p) which predicts the dependent variable.
            p (array_like or float): Constant parameters to be passed to the function.
            x (list): Values of the abscissa at which to construct a grid.
            y (list): Values of the oordinate at which to construct a grid.
            ax (Axes): Optional existing axes on which to plot.
            log (bool): Flag whether or not to plot log scale.
        
        Returns:
            ax (Axes): Final axes object containing the plot.
            handles (list): Handles of the plotted elements.
    """
    ## make axes if none given
    if ax == None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure
    M = len(x)
    N = len(y)
    Z = np.zeros((M,N))
    for i in range(M):
        for j in range(N):
            Z[i,j] = f(x[i],y[j],*p)
    handles = ()
    v = [x,y,Z.T]
    v = [np.log(v[i]) if log[i] else v[i] for i in range(3)]
    pltargs = {'levels':20,'cmap':'jet'}
    pltargs.update(kwargs)
    if filled:
        h = ax.contourf(*v,**pltargs)
    else:
        h = ax.contour(*v,**pltargs)
    handles += (h,)
    if new_cbar:
        cbar = fig.colorbar(h)
    else:
        cbar = ax.collections[-1].colorbar
    
    plt.close()
    return ax, handles, cbar, Z

